// Eric Lundell
// 17 Dec 2010
// Project 2, CS 301 Assembly Language
// Link with abeep.asm
//
// This program will read in a specifc format file and beep the internal speaker at a defined pitch.
// File format is described in the files themselves and the README.
//
// To compile this program most easily, use the shell Proj.sh.  This will compile both the .cpp and .asm file.
// To run you must use sudo:
//
// 	sudo ./abeep musicfile.txt
//

#include <fcntl.h>     // for ioctl
#include <sys/ioctl.h> // " "
#include <linux/kd.h>  // for KDMKSOUND

#include <iostream>
#include <vector>
#include <fstream>
#include <string>
using namespace std;

#include <stdio.h>
#include <string.h> // for error handling
#include <errno.h>  // for error handling
#include <stdlib.h> // for exit(1)

#define _CLOCK 1193180 // (0x1234DC)
vector<char> _music;  // a global that holds read in music

bool DEBUG = false;  // Extra messaging for debugging if true


extern "C" void abeep_asm(void);
extern "C" int fda_asm;  // avoid function parameter calls - less error for now

float makeToneNum(char note, char octave, char pitch)
{
    int tone = 0;
    float oct = 0;
    float base = 1000;
    float freq = 0;

    // note that notes are ms long
    switch(note)
    {
        case 'w':  base = 1.000 * base; break; // whole
        case '2':  base = 0.750 * base; break; // dotted half
        case 'h':  base = 0.500 * base; break; // half
        case '4':  base = 0.330 * base; break; // dotted quarter
        case 'q':  base = 0.250 * base; break; // quarter
        case '8':  base = 0.188 * base; break; // dotted eighth
        case 'e':  base = 0.125 * base; break; // eighth
	case 's':  base = 0.063 * base; break; // sixteen
        default: base = 0;
		cout << " WARNING: " << note << " is not a proper note.\n";
		cout << " This note will not sound." << endl; 
		break; // if no or wrong note given, no tone is sounded

    }

    tone = ((int)base) << 16;

    // Octave ranges
    switch(octave)
    {
        case '1': oct = 0.33; break;  // 2 octaves below Middle C
	case '2': oct = 0.5; break;  // 1 octaves below Middle C
	case '3': oct = 1; break;  // Middle C
	case '4': oct = 2; break;  // 1 octaves above Middle C
	case '5': oct = 3; break;  // 2 octaves above Middle C
        default: if(pitch == '0') break;  // if a rest, ignore octave
		oct = 1; 
		cout << " WARNING: " << octave << " is not a proper octave.\n";
		cout << " This note will not sound." << endl;
		tone = 0;	
		break;

    }

    // Middle C range
    switch(pitch)
    { 
        case '0': freq = 0;   break; // This is used to represent rests
        case 'a': freq = 220; break; // A - 220 Hz
        case 'A': freq = 233; break; // A# - 233 Hz
        case 'b': freq = 241; break; // B - 241 Hz
        case 'B': // B# = c
        case 'c': freq = 261; break; // C - 261 Hz
        case 'C': freq = 277; break; // C# - 277 Hz
        case 'd': freq = 294; break; // D - 294 Hz
        case 'D': freq = 311; break; // D# - 311 Hz
        case 'e': freq = 330; break; // E - 330 Hz
        case 'E': // E# = f
        case 'f': freq = 349; break; // F - 349 Hz
        case 'F': freq = 370; break; // F# - 370 Hz 
        case 'g': freq = 392; break; // G - 392 Hz
        case 'G': freq = 415; break; // G# - 415 Hz
	default: freq = 1; // Can't have zero because of later division
		cout << " WARNING: " << pitch << " is not a proper pitch.\n";
		cout << " This note will not sound." << endl;
		tone = 0;
		break;  
    }

    // freq == 0 means a rest is being played, so same time length but no tone
    if(freq == 0) tone = tone | (0x0000FFFF & 0x0 );

    tone = tone | (0x0000FFFF & int( _CLOCK / (freq * oct)) );

    if(DEBUG) cout << " TONE: " << tone << endl;

    return tone;
}

void readMusic(int argc, char** argv)
{
        if(argc != 2)
        {
            cout << " ERROR: Need one music file to read in music!\n";
            cout << " Will sound basic beep instead.\n\n";
           
            _music.push_back('w');
            _music.push_back('3');
            _music.push_back('c');  
        }
	else
	{
	    ifstream musicFile;
	    musicFile.open(argv[1]);

	    if(!musicFile.is_open())
	    {
                 cout << "  ERROR: could not open music file.\n";
		 exit(1);
	    }

	    char tChar;
            string line;

	    while(!musicFile.eof())
	    {
		 getline(musicFile, line);

		 for(int i = 0; i < line.size(); i++)
		 { 
			if(DEBUG) cout << " Reading music file: now at " << line.at(i) << endl;
   		 	if(line.at(i) == '\n') 
				break; // time for next line
			else if(line.at(i) == ' '); // skip ahead
                        else if(line.at(i) == '#') 
				break; // to the next line
		 	else 
				_music.push_back( line.at(i) );
		 }
            }

	    musicFile.close();
	}
}


int main(int argc, char** argv)
{
    int fd  = -2;
    int sound = 0;
    int ioctlCheck = -1;

    //fd = open(device, 1);
    // call assembly to open device
    abeep_asm();
    if(DEBUG) cout << "abeep: " << strerror(errno) << " fd: " << fda_asm << endl;
    fd = fda_asm;

    // Read in music file
    // vector<char> _music made global
    //void readMusic(int argc, char** argv)
    readMusic(argc, argv);
    
    for(int i = 0; i < _music.size(); i += 3)
    {
        //float makeToneNum(char note, char octave, char pitch)
        sound = makeToneNum(_music[i], _music[i+1], _music[i+2]);
        ioctlCheck = ioctl(fd, KDMKTONE, sound);  
        if(DEBUG) cout << "ioctlCheck: " << ioctlCheck << " Ioctl: " << strerror(errno) << endl;
        // ioctl is threaded and returns instantly.  Thus all the ioctl calls
	// will be overwritten but the last one.  To solve this, call usleep for the 
	// length of the note just called.   
        usleep( 2000*((sound&0xFFFF0000) >> 16) ); 
    }

    // close file
    close(fd);
    if(DEBUG) cout << "Close: " << strerror(errno) << endl;

    return 0;
 
}



